// Contour Bombing Cave Generation Algorithm
//
// License notice: You can do anything with this source code and algorithm itself, except:
//	- You cannot claim it as your own, even if you derive your work from this source code and change it
//  - It would be welcome to put me on a credit/thank you/'anything you want' page, but you are not obligated to do so.
// If you decide to credit me, a text stating: Special thanks to darkgnostic, would be enough.
//
#include "stdafx.h"
#include <vector>
#include <thread>
#include "rang.h"

using namespace rang;
#undef min
#undef max

#include <algorithm>


#define MAP_WIDTH		80
#define MAP_HEIGHT		35

void move_to( HANDLE hStdout, int x, int y) {
	
	COORD destCoord;
	//position cursor at start of window
	destCoord.X = x;
	destCoord.Y = y;
	SetConsoleCursorPosition(hStdout, destCoord);
}
std::vector<int> tile_map;

int random_gen_get_i(unsigned i) {
	return rand() % i;
}

int random_gen_range(unsigned min, unsigned max) {
	return (min + (rand() % (int)(max - min + 1)));
}

void set_tile(int x, int y, int value = 1) {
	if (0 <= x && x < MAP_WIDTH && 0 <= y && y < MAP_HEIGHT)
		tile_map[x + y * MAP_WIDTH] = value;
}

// Untested, but simple enough it should at least be close to reality...
void print_map( bool draw_space = false ) {

	//initialize objects for cursor manipulation
	HANDLE hStdout;

	hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
	unsigned i = 0;
	for (; i < tile_map.size(); i++) {

		if (tile_map[i] || draw_space ) {
			move_to(hStdout, i%MAP_WIDTH, i / MAP_WIDTH);

			// 2 green
			SetConsoleTextAttribute(hStdout, 1 + tile_map[i]);
			printf("%c", tile_map[i] ? '#' : ' ');
		}
	}
	move_to(hStdout, i%MAP_WIDTH, i / MAP_WIDTH);
}


bool build_cavern( bool use_borders = true) {

	int min_x = 0, min_y = 0, max_x = MAP_WIDTH, max_y = MAP_HEIGHT;

	// create list of tile candidates
	std::vector<std::pair<int, int>> candidates;

	int k = 0;
	for (auto& tile : tile_map ) {
		if (tile ) {
			candidates.emplace_back(std::make_pair(k% MAP_WIDTH, k / MAP_WIDTH ));
		}
		k++;
	}

	// shuffle them all
	for ( unsigned p = 0; p < candidates.size(); p++)
		std::swap(candidates[p], candidates[random_gen_get_i(candidates.size())]);

	unsigned iteration_number = (unsigned)((float)candidates.size() * 4.8f);

	for (unsigned i = 0; i < iteration_number; i++) {

		unsigned random_offset = 0;
		auto tile_variation = 0;

		// 1/3 chance that we will use as a bombing point one of the last 15 positions
		if (random_gen_get_i(3) == 0) {
			random_offset = random_gen_range(candidates.size() - 15, candidates.size() - 1);
			tile_variation = 1;
		}
		else {
			// otherwise use lower half of remaining tiles
			random_offset = random_gen_get_i(candidates.size()/2);
			tile_variation = 2;
		}

		// check boundaries
		if (random_offset >= candidates.size())
			random_offset = candidates.size() - 1;

		int x = candidates[random_offset].first;
		int y = candidates[random_offset].second;

		int map_width = MAP_WIDTH;
		int map_height = MAP_HEIGHT;
		bool use_borders = true;

		// we will use bombs of radius 1 mostly with smaller chance(1 / 20)
		// that radius will be of size 2 
		int bomb_radius = random_gen_get_i(20) != 0 ? 1 : 2;

		// bomb
		for (int i = std::max(0, x - bomb_radius - 1); i < std::max(map_width, x + bomb_radius); i++) {
			for (int j = std::max(0, y - bomb_radius - 1); j < std::max(map_height, y + bomb_radius); j++) {
				if ((i - x)*(i - x) + (j - y)*(j - y) < bomb_radius * bomb_radius + bomb_radius) {

					int map_offset = i + j * MAP_WIDTH;

					if (0 <= map_offset && map_offset < (int)tile_map.size())	{
					
						if (use_borders) {
							if (i < min_x) continue;
							if (i > max_x) continue;
							if (j < min_y) continue;
							if (j > max_y) continue;
						}
						
						auto& data = tile_map[map_offset];

						// if we have at least one tile bombed on screen
						// push those coordinates to candidate list
						if (tile_map[map_offset] != tile_variation) 	{
							set_tile(i, j, tile_variation);
							candidates.emplace_back(std::make_pair(i, j));
						}
					}
				}
			}
		}
		
		// erase our cell, it is re-added in bombing loop above, if at least one tile is changed.
		candidates.erase(candidates.begin() + random_offset);
	}

	candidates.clear();

	return true;
}




int _tmain(int argc, _TCHAR* argv[])
{
	srand(time(0));

	for (int i = 0; i < 10; i++) {

		tile_map.clear();
		tile_map.resize(MAP_HEIGHT*MAP_WIDTH);

		// line 
		/*for (int x = 25; x < 60; x++) {
			set_tile(x, 13);
			set_tile(x, 14);
			set_tile(x, 15);
		}

		// T
		/*for (int x = 25; x < 60; x++) {
			set_tile(x, 15);
		}

		for (int x = 15-10; x < 16+10; x++) {
			set_tile(25, x);
		}
		//*/

		// test wave
		
		/*for (int x = 25; x < 60; x++) {
			set_tile(x, 15+sin(x/4)*3);
		}

		for (int x = 15 - 10; x < 16 + 10; x++) {
			set_tile(25, x);
		}//*/

		// wave
		for (int x = 25; x < 60; x++) {
			set_tile(x, 8 + sin(x / 4) * 3);
			set_tile(x, 24 + sin(x / 4) * 3);
		}
		for (int x = 8; x < 24; x++) {
			set_tile(25, x);
			set_tile(59, x+3);
		}
		print_map(true);
		getchar();
		build_cavern();
		print_map();
		getchar();
	}


	
	
	return 0;
}

